#include <l4/sys/types.h>
#include <l4/sys/syscalls.h>
#include <l4/sys/ipc.h>
#include <l4/sys/kernel.h>
#include <l4/sys/kdebug.h>
#include <l4/rmgr/librmgr.h> 

#include <stdio.h>
#include <assert.h>

extern l4_threadid_t rmgr_pager_id;

int main (void)
{
  l4_umword_t phys_addr;
  int err;
  l4_msgdope_t dope;
  l4_snd_fpage_t fp;

  printf ("rmgrtest running\n");

  rmgr_init();

  /* reserve main-memory page between 4 MB and 8 MB, then free it again. */
  phys_addr = rmgr_reserve_mem (L4_PAGESIZE, L4_LOG2_PAGESIZE, 
				0, 0x401000, 0x800000);
  assert (phys_addr != -1);

  printf("reserved page at phys_addr=%d\n", phys_addr);

  /* map that page at virt. address 8 MB */
  err = l4_i386_ipc_call (rmgr_pager_id, 
			  L4_IPC_SHORT_MSG, phys_addr, 0,
			  L4_IPC_MAPMSG (0x800000, L4_LOG2_PAGESIZE), 
			  &fp.snd_base, 
                          &fp.fpage.fpage,
                          L4_IPC_NEVER,
                          &dope);

  assert (err == 0);
  assert (dope.md.fpage_received);

  enter_kdebug ("flexpage received");

  /* flush page again */
  rmgr_free_page (phys_addr);

  enter_kdebug ("flexpage flushed");

  return 0;
}
