#include <l4/sys/types.h>
#include <l4/sys/syscalls.h>
#include <l4/sys/ipc.h>
#include <l4/sys/kernel.h>
#include <l4/sys/kdebug.h>
#include <l4/rmgr/librmgr.h> 

#include <stdio.h>
#include <assert.h>

#define MAP_ADDRESS	0x40000000

extern l4_threadid_t rmgr_pager_id;

int main (void)
{
  l4_umword_t phys_addr, addr, map_addr;
  int err;
  l4_msgdope_t dope;
  l4_snd_fpage_t fp;

  printf ("rmgrtest running\n");

  rmgr_init();

  /* reserve main-memory pages. */
  phys_addr = rmgr_reserve_mem (2*L4_SUPERPAGESIZE, L4_LOG2_SUPERPAGESIZE,
				0, 0, 0);
  assert (phys_addr != -1);

  printf("reserved page at phys_addr=%08x, mapping to %08x\n", 
      phys_addr, MAP_ADDRESS);

  /* map pages */
  for (addr = phys_addr, map_addr = MAP_ADDRESS;
       addr < phys_addr + 2*L4_SUPERPAGESIZE; 
       addr += L4_SUPERPAGESIZE, map_addr += L4_SUPERPAGESIZE)
    {
      err = l4_i386_ipc_call (rmgr_pager_id,
			      L4_IPC_SHORT_MSG, 
			        addr | 1, L4_LOG2_SUPERPAGESIZE << 2,
			      L4_IPC_MAPMSG (map_addr, L4_LOG2_SUPERPAGESIZE),
		      	      &fp.snd_base, 
	      		      &fp.fpage.fpage,
      			      L4_IPC_NEVER,
			      &dope);
      assert (err == 0);
      assert (dope.md.fpage_received);
    }

  enter_kdebug ("flexpages received");

  for (addr = phys_addr, map_addr = MAP_ADDRESS;
       addr < phys_addr + 2*L4_SUPERPAGESIZE; 
       addr += L4_SUPERPAGESIZE, map_addr += L4_SUPERPAGESIZE)
    {
      printf("unmapping superpage at %08x\n", map_addr);
      l4_fpage_unmap(l4_fpage(map_addr, L4_LOG2_SUPERPAGESIZE, 0, 0),
			      L4_FP_ALL_SPACES|L4_FP_FLUSH_PAGE);
      enter_kdebug("flexpage flushed");
    }

  enter_kdebug ("flexpages flushed");
  
  return 0;
}
