/* Extract symbols information from ELF binary and tell Fiasco where 
 * to find it */

#include <stdio.h>
#include <l4/sys/types.h>
#include "exec.h"
#include "elf.h"
#include "globals.h"
#include "init.h"
#include "memmap.h"
#include "symbols.h"

void
extract_symbols(l4_addr_t elf_image, unsigned task_no,
		l4_addr_t *from_sym, l4_addr_t *to_sym)
{
  int i;
  Elf32_Ehdr *ehdr = (Elf32_Ehdr*)elf_image;
  Elf32_Shdr *sh_sym, *sh_str;
  unsigned sz_str, sz_sym;

  if (l4_version != VERSION_FIASCO)
    return;

  for (i=0; i<ehdr->e_shnum; i++)
    {
      Elf32_Sym *sym_symtab;
      const char *sym_strtab;

      sh_sym = (Elf32_Shdr*)(elf_image + ehdr->e_shoff
				       + i*ehdr->e_shentsize);
      if (sh_sym->sh_type == SHT_SYMTAB)
	{
	  unsigned num_symtab;
	  Elf32_Sym *sym;
	  unsigned size, len;
	  unsigned pages;
	  char *str;
	  l4_addr_t addr, syms;

	  sh_str = (Elf32_Shdr*)(elf_image + ehdr->e_shoff
					   + sh_sym->sh_link*ehdr->e_shentsize);
	  sz_str     = sh_str->sh_size;
	  sym_strtab = (char*)(elf_image + sh_str->sh_offset);
	  sz_sym     = sh_sym->sh_size;
	  sym_symtab = (Elf32_Sym*)(elf_image + sh_sym->sh_offset);
	  num_symtab = sz_sym/sizeof(Elf32_Sym);
	  size       = 0;

	  // first determine size of symbol table
	  for (i=0, sym=sym_symtab; i<num_symtab; i++, sym++)
	    {
	      const char *s_name = sym_strtab + sym->st_name;

	      // ignore something special
	      if (  (sym->st_shndx >= SHN_LORESERVE)
		  ||(sym->st_value == 0)
		  ||(*s_name == '\0')
		  ||(!memcmp(s_name, "Letext", 6))
		  ||(!memcmp(s_name, "_stext", 6)))
		continue;

	      len = strlen(s_name);
	      if (len>100)
		len = 100;
	      size += 11+len+1;
	    }

	  pages = l4_round_page(size)/L4_PAGESIZE;
	  if (0 == (addr = syms = find_free_chunk(pages, 1)))
	    {
	      printf("      No space for symbols left\n");
	      return;
	    }
	  str  = (char*)addr;

	  // now copy symbols into area
	  for (i=0, sym=sym_symtab; i<num_symtab; i++, sym++)
	    {
	      const char *s_name = sym_strtab + sym->st_name;

	      // ignore something special
	      if (  (sym->st_shndx >= SHN_LORESERVE)
		  ||(sym->st_value == 0)
		  ||(*s_name == '\0')
		  ||(!memcmp(s_name, "Letext", 6))
		  ||(!memcmp(s_name, "_stext", 6)))
		continue;

	      str += snprintf(str, 11+100+1, "%08x   %s\n",
			      sym->st_value, s_name);
	    }

	  // mark pages as reserved
	  for (i=0; i<pages; i++, addr+=L4_PAGESIZE)
	    memmap_alloc_page(addr, O_RESERVED);

	  // now publish symbols to fiasco
	  asm ("int $3 ; cmpb $30,%%al" : : "a"(syms), "b" (task_no), "c" (1));

	  *from_sym = syms;
	  *to_sym   = syms+L4_PAGESIZE*pages;

	  return;
	}
    }

  return;
}
